unit uDemandas;

{$MODE Delphi}

(*** DEFINE EL PADRE DE TODAS LAS DEMANDAS *****)
interface

uses
  unodos, uActorNodal, uFechas, xMatDefs, SysUtils,
  usimplex, Classes,
  ucosa, uCosaConNombre,
  uConstantesSimSEE, uFuentesAleatorias,
  uglobs, Math,
  uauxiliares, uFichasLPD;

type
  TDemanda = class(TActorUniNodal)
  public
    falla_profundidad: TDAOfNReal; //profundidad de los escalones en p.u.
    falla_costo_0, falla_costo_indexado: TDAOfNReal; // costo de falla en USD/MWh
    costos: array of TDAOfNReal; //Costos resultantes del paso para cada
    //escalon para cada poste.
    //costos[iEscalon][iPoste] = fallas[iEscalon][iPoste] * durPos[iPoste] * falla_costo_indexado[iEscalon]

    PPA_PreSorteos: TDAOfNREal; // potencias por poste pre-sorteos.
    PPa: TDAOfNReal; // potencias por poste actuales (pos-sorteos).

    fallas: array of TDAOfNReal; // potencias de cada escaln fallas en cada poste
    // hay un vector por cada escaln y en cada vector
    // estn las potencias despachadas de ese escaln
    // en cada poste.

    fuente: TFuenteAleatoria;
    nombreBorne: string;
    numeroBorne: integer;
    SumarEnergiaHr: boolean;
    //Indica si debe sumar el valor de la bornera fuente a la potencia o debe utilizarlo como coeficiente P*(1+borne)
    icf_Fuente: TFuenteAleatoria;
    icf_NombreBorne: string;
    icf_kBorne: integer;
    icf_Valor: NReal;

    iaRand: NReal; // factor aleatorio de la demana  (1 + fuente)

    PHoraria_: TDAOfNReal;

    // auxiliar, cada demanda carga el detalle del paso y luego se postiza
    idxHorasPostizadas: TDAOfNInt;

    constructor Create(capa: integer; nombre: string; nacimiento, muerte: TFecha;
      lpdUnidades: TFichasLPD; nodo: TNodo; falla_profundidad, falla_costo: TDAOfNReal;
      fuente: TFuenteAleatoria; nombreBorne: string; icf_Fuente: TFuenteAleatoria;
      icf_NombreBorne: string; SumarEnergiaHr: boolean);

    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;

    function NEscalonesDeFalla: integer;
    procedure PrepararMemoria(globs: TGlobs); override;
    procedure PrepararPaso_ps; override;

    procedure opt_cargue(s: TSimplex); override;

    function demandaPromedio(fechaIni, fechaFin: TFecha): NReal; virtual; abstract;
    function demandaMaxima(fechaIni, fechaFin: TFecha): NReal; virtual; abstract;

    procedure opt_nvers(var ivar, ivae, ires: integer); override;
    procedure opt_fijarRestriccionesDeCaja(s: TSimplex); override;
    procedure opt_leerSolucion(s: TSimplex); override;
{$IFDEF SPXCONLOG}
    procedure spx_NombrarVariables(s: TSimplex); override;
{$ENDIF}
    function getNombreVar(ivar: integer; var nombre: string): boolean; override;
    function getNombreRes(ires: integer; var nombre: string): boolean; override;

    procedure PubliVars; override;
    procedure dump_Variables(var f: TextFile; charIndentacion: char); override;
    procedure prepararPaso_as; override;
    procedure Free; override;
  end;




(* Escalones de falla por defecto
( Uruguay-invierno 2006 )
escf [pu]  0.05  0.075  0.075  0.8
cvf [USD/MWh]  250  400  1200  2000
*)
(* Retorna un array con las profundidades de la falla por defecto *)
function ProfundidadEscalonesDeFallaPorDefecto: TDAOfNReal;
(* Retorna un array con los costos de falla por defecto *)
function CostoEscalonesDeFallaPorDefecto: TDAOfNReal;

implementation


constructor TDemanda.Create(capa: integer; nombre: string;
  nacimiento, muerte: TFecha; lpdUnidades: TFichasLPD; nodo: TNodo;
  falla_profundidad, falla_costo: TDAOfNReal; fuente: TFuenteAleatoria;
  nombreBorne: string; icf_Fuente: TFuenteAleatoria;
  icf_NombreBorne: string; SumarEnergiaHr: boolean);
begin
  inherited Create(capa, nombre, nacimiento, muerte, lpdUnidades, nodo);
  self.falla_profundidad := copy(falla_profundidad, 0, length(falla_profundidad));
  self.falla_costo_0 := copy(falla_costo);
  self.fuente := fuente;
  self.nombreBorne := nombreBorne;
  self.icf_Fuente := icf_Fuente;
  self.icf_NombreBorne := icf_NombreBorne;
  self.SumarEnergiaHr := SumarEnergiaHr;
end;

(* Escalones de falla por defecto
( Uruguay-invierno 2006 )
escf [pu]  0.05  0.075  0.075  0.8
cvf [USD/MWh]  250  400  1200  2000
*)
function ProfundidadEscalonesDeFallaPorDefecto: TDAOfNReal;
var
  a: TDAOfNReal;
begin
  setlength(a, 4);
  a[0] := 0.05;
  a[1] := 0.075;
  a[2] := 0.075;
  a[3] := 0.8;
  Result := a;
end;

function CostoEscalonesDeFallaPorDefecto: TDAOfNReal;
var
  a: TDAOfNReal;
begin
  setlength(a, 4);
  a[0] := 250.0;
  a[1] := 400.0;
  a[2] := 1200.0;
  a[3] := 2000.0;
  Result := a;
end;

function TDemanda.NEscalonesDeFalla: integer;
begin
  Result := length(falla_profundidad);
end;

procedure TDemanda.PrepararMemoria(globs: TGlobs);
var
  iescalon: integer;
begin
  inherited PrepararMemoria(globs);

  setlength(PPA, globs.NPostes);
  setlength(PPA_PreSorteos, globs.NPostes);
  // creamos el espacio para el resultado del despacho de las fallas y sus costos
  setlength(fallas, length(falla_profundidad));
  SetLength(costos, length(falla_profundidad));
  for iescalon := 0 to high(fallas) do
  begin
    setlength(fallas[iescalon], globs.NPostes);
    SetLength(costos[iescalon], globs.NPostes);
  end;

  setlength(falla_costo_indexado, length(fallas));

  if fuente <> nil then
    numeroBorne := fuente.idBorne(nombreBorne);
  if icf_Fuente <> nil then
    icf_kBorne := icf_Fuente.IdBorne(icf_NombreBorne);

  setlength(PHoraria_, ceil(globs.HorasDelPaso));
  setlength(idxHorasPostizadas, ceil(globs.HorasDelPaso));

  // por ahora hago esto para que la primer demanda que aparezca
  // sea la principal.
  if globs.GetDemandaPrincipal = nil then
    globs.SetDemandaPrincipal(Self);
end;

procedure TDemanda.prepararPaso_ps;
var
  iposte: integer;
begin
  if icf_Fuente <> nil then
  begin
    icf_Valor := icf_Fuente.Bornera[icf_kBorne];
    for iposte := 0 to high(self.falla_costo_indexado) do
      self.falla_costo_indexado[iposte] := self.falla_costo_0[iposte] * icf_Valor;
  end
  else
  begin
    icf_Valor := 1;
    for iposte := 0 to high(self.falla_costo_indexado) do
      self.falla_costo_indexado[iposte] := self.falla_costo_0[iposte];
  end;

  if fuente <> nil then
  begin
    if SumarEnergiaHr then
    begin //si es usado como incremento de la demanda
      iaRand := 1.0;
      for iposte := 0 to high(PPA) do
        PPA[iposte] := PPA_PreSorteos[iposte] + fuente.bornera[numeroBorne];
    end
    else
    begin //si es usado como factor de alteracin de la demanda
      iaRand := 1 + fuente.bornera[numeroBorne];
      if iaRand < 0 then
        iaRand := 0.0;
      for iposte := 0 to high(PPA) do
        PPA[iposte] := PPA_PreSorteos[iposte] * iaRand;
    end;
  end
  else
  begin
    iaRand := 1.0;
    for iposte := 0 to high(PPA) do
      PPA[iposte] := PPA_PreSorteos[iposte];
  end;
end;

procedure TDemanda.opt_cargue(s: TSimplex);
var
  iposte, iescalon: integer;
  ibaseres: integer;
begin
{$IFDEF SPXCONLOG}
  spx_NombrarVariables(s);
{$ENDIF}

  ibaseres := nodo.ires;
  // Carga el trmino constante de las restricciones de demanda
  for iposte := 0 to high(PPa) do
    s.acum_e(ibaseres + iposte, s.nc, -PPa[iposte]);

  // Ahora corresponde cargar las mquinas de falla
  for iescalon := 0 to high(falla_profundidad) do
    for iposte := 0 to globs.NPostes - 1 do
    begin
      // cargamos la potencia en la restriccin del nodo
      s.pon_e(ibaseres + iposte, ivar + iescalon * globs.NPostes + iposte, 1);
      // cargamos el coeficiente del costo en la funcin objetivo
      s.pon_e(s.nf, ivar + iescalon * globs.NPostes + iposte,
        -falla_costo_indexado[iescalon] * globs.durpos[iposte]);
    end;
end;

constructor TDemanda.Create_ReadFromText(f: TArchiTexto);
begin
  inherited Create_ReadFromText(f);
  f.IniciarLecturaRetrasada;
  f.rd('falla_profundidad', falla_profundidad);
  f.rd('falla_costo', falla_costo_0);
  f.rdReferencia('fuente', TCosa(fuente), Self);
  f.rd('nombreBorne', nombreBorne);

  if f.Version < 8 then
  begin
    icf_Fuente := nil;
    icf_NombreBorne := '';
  end
  else
  begin
    f.rdReferencia('indiceCostoDeFalla', TCosa(icf_Fuente), Self);
    f.rd('borneIndiceCostoDeFalla', icf_NombreBorne);
  end;

  if f.Version >= 74 then
    f.rd('SumarEnergiaHr', SumarEnergiaHr)
  else
    SumarEnergiaHr := False;
end;

procedure TDemanda.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wr('falla_profundidad', falla_profundidad, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALESPU);
  f.wr('falla_costo', falla_costo_0, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);
  f.wrReferencia('fuente', fuente);
  f.wr('nombreBorne', nombreBorne);
  f.wrReferencia('indiceCostoDeFalla', icf_Fuente);
  f.wr('borneIndiceCostoDeFalla', icf_NombreBorne);
  f.wr('SumarEnergiaHr', SumarEnergiaHr);
end;

procedure TDemanda.PrepararPaso_as;
var
  iposte: integer;
  hora: integer;
  nHorasDelPoste: integer;
  EnergiaDelPoste: NReal;
  jhora: integer;
  k: integer;
begin
  if not globs.SalaMinutal then
  begin
    if self = globs.GetdemandaPrincipal then
    begin
      // inicializamos el ndice
      for k := 0 to high(idxHorasPostizadas) do
        idxHorasPostizadas[k] := k;
      if globs.PostesMonotonos then
        QuickSort_Decreciente(PHoraria_, idxHorasPostizadas);
      hora := 0;
      for iposte := 0 to globs.NPostes - 1 do
      begin
        nHorasDelPoste := round(globs.Durpos[iposte]);
        EnergiaDelPoste := 0;

        for jhora := 0 to nHorasDelPoste - 1 do
        begin
          EnergiaDelPoste := EnergiaDelPoste + PHoraria_[hora];

          globs.kPosteHorasDelPaso[idxHorasPostizadas[hora]] := iposte;
          Inc(hora);
        end;
        PPA_PreSorteos[iposte] := EnergiaDelPoste / nHorasDelPoste;
      end;
    end
    else
    begin
      vclear(PPA_PreSorteos);
      for hora := 0 to high(PHoraria_) do
      begin
        iposte := globs.kPosteHorasDelPaso[hora];
        PPA_PreSorteos[iposte] := PPA_PreSorteos[iposte] + PHoraria_[hora];
      end;
      for iposte := 0 to high(PPA) do
      begin
        nHorasDelPoste := round(globs.Durpos[iposte]);
        PPA_PreSorteos[iposte] := PPA_PreSorteos[iposte] / nHorasDelPoste;
      end;
    end;
  end
  else
  begin
    PPA_PreSorteos[0]:= PHoraria_[0];
  end;
end;

procedure TDemanda.Free;
var
  i: integer;
begin
  SetLength(falla_profundidad, 0);
  SetLength(falla_costo_0, 0);
  SetLength(falla_costo_indexado, 0);
  for i := 0 to high(costos) do
    SetLength(costos[i], 0);
  SetLength(costos, 0);
  SetLength(PPA_PreSorteos, 0);
  SetLength(PPa, 0);
  for i := 0 to high(fallas) do
    SetLength(fallas[i], 0);
  SetLength(fallas, 0);
  SetLength(PHoraria_, 0);
  SetLength(idxHorasPostizadas, 0);
  inherited Free;
end;

procedure TDemanda.opt_nvers(var ivar, ivae, ires: integer);
begin
  Self.ivar := ivar;
  // tengo para cada escaln de falla una variable por poste
  ivar := ivar + NEscalonesDeFalla * globs.NPostes;
end;

procedure TDemanda.opt_fijarRestriccionesDeCaja(s: TSimplex);
var
  iescalon, iposte: integer;
  Fpmax: NReal;
begin
  for iescalon := 0 to high(falla_profundidad) - 1 do
    // Le fijamos la potencia mxima de cada escaln en cada poste
    for iposte := 0 to globs.NPostes - 1 do
    begin
      Fpmax := PPa[iposte] * Self.falla_profundidad[iescalon];
      s.cota_sup_set(ivar + iescalon * globs.NPostes + iposte, Fpmax);
    end;

  // al ltimo escaln le ponemos un poquitito ms para que siempre sea capaz de
  // satisfacer la demanda sin iportar los errores numricos
  iescalon := high(falla_profundidad);
  // Le fijamos la potencia mxima de cada escaln en cada poste
  for iposte := 0 to globs.NPostes - 1 do
  begin
{$IFDEF HOLGURA_FALLA}
    Fpmax := PPa[iposte] * Self.falla_profundidad[iescalon] * 1.002;
{$ELSE}
    Fpmax := PPa[iposte] * Self.falla_profundidad[iescalon];
{$ENDIF}
    s.cota_sup_set(ivar + iescalon * globs.NPostes + iposte, Fpmax);
  end;

{$IFDEF GATILLOS_CAMBIOVAR}
  // Gatillamos los cambios de variables para dar con una solucin fatible
  // desde el inicio
  for iescalon := 0 to high(falla_profundidad) do
    for iposte := 0 to globs.NPostes - 1 do
      s.GatillarCambioVarCotaSup(ivar + iescalon * globs.NPostes + iposte);
{$ENDIF}
end;

procedure TDemanda.opt_leerSolucion(s: TSimplex);
var
  cvf, fallaIJ: NReal;
  iescalon, iposte: integer;
  m: NReal;
begin
  // recuperamos los valores de Potencia de cada escaln de falla en cada poste.
  costoDirectoDelPaso := 0;
  vclear(P);
  for iescalon := 0 to high(falla_profundidad) do
  begin
    cvf := falla_costo_indexado[iescalon];
    for iposte := 0 to globs.NPostes - 1 do
    begin
      fallaIJ := s.xval(ivar + iescalon * globs.NPostes + iposte);
      P[iposte] := P[iposte] + fallaIJ;
      fallas[iescalon][iposte] := fallaIJ;
      m := fallaIJ * cvf * globs.DurPos[iposte];
      costos[iescalon][iposte] := m;
      costoDirectoDelPaso := costoDirectoDelPaso + m;
    end;
  end;
  for iposte := 0 to globs.NPostes - 1 do
    P[iposte] := P[iposte] - PPA[iposte];
end;

{$IFDEF SPXCONLOG}
procedure TDemanda.spx_NombrarVariables(s: TSimplex);
var
  iposte, iescalon: integer;
begin
  for iescalon := 0 to high(falla_profundidad) do
    for iposte := 0 to globs.NPostes - 1 do
      s.set_NombreVar(ivar + iescalon * globs.NPostes + iposte,
        nombre + '_[MW]F' + IntToStr(iescalon + 1) + 'p' + IntToStr(iposte + 1));
end;

{$ENDIF}

function TDemanda.getNombreVar(ivar: integer; var nombre: string): boolean;
var
  iEscalon: integer;
  iPoste: integer;
begin
  if (ivar >= self.ivar) and (ivar < self.ivar + NEscalonesDeFalla * globs.NPostes) then
  begin
    iEscalon := (ivar - self.ivar) div (globs.NPostes) + 1;
    iPoste := (ivar - self.ivar) mod (globs.NPostes) + 1;
    nombre := self.nombre + '_[MW]F' + IntToStr(iescalon) + 'p' + IntToStr(iposte);
    Result := True;
  end
  else
    Result := False;
end;

function TDemanda.getNombreRes(ires: integer; var nombre: string): boolean;
begin
  Result := False;
end;

procedure TDemanda.PubliVars;
var
  iescalon: integer;
begin
  inherited PubliVars;
  PublicarVariableVR('PD', '[MW]', 6, 1, PPA, True, True);
  for iescalon := 0 to high(fallas) do
    PublicarVariableVR('PF' + IntToStr(iescalon + 1), '[MW]', 6, 1,
      fallas[iescalon], True, True);
  for iescalon := 0 to high(fallas) do
    PublicarVariableVR('Costo' + IntToStr(iescalon + 1), '[USD]', 8, 1,
      costos[iescalon], True, True);
end;

procedure TDemanda.dump_Variables(var f: TextFile; charIndentacion: char);
var
  i: integer;
begin
  inherited dump_Variables(f, charIndentacion);
  for i := 0 to high(PPA) do
    writeln(f, charIndentacion, 'PPA[' + IntToStr(i + 1) + ']= ',
      FloatToStrF(PPA[i], ffFixed, 10, 3));
  writeln(f);
end;



end.
